"""
decorator.py

Contains decorator definitions for use.

Valid uses of @service are

  @service('id')
  def func(arg): pass

  @service('id')
  class I(object): pass

Valid uses of @inject are

  @inject(['id1','id2'])
  def func(a,b): pass

  class O(object):
    @inject(['id1','id2'])
    def __init__(self,a,b): pass
"""

import inspect
from di import ServiceStore

class Servicer(object):
    """@service decoration"""
    def __init__(self, id, store=ServiceStore, ns='default'):
        self.id = id
        self.store = store
        self.namespace = ns

    def __call__(self, f):
        if inspect.isclass(f):
            self.store.add_service(self.id, f(), to_namespace=self.namespace)
        else:
            self.store.add_service(self.id, f, to_namespace=self.namespace)

        return f

class Injector(object):
    """@inject decoration"""
    def __init__(self, list_of_service_ids, store=ServiceStore, ns='default'):
        self.ids = list_of_service_ids
        self.store = store
        self.namespace = ns

    # TODO: figure out a better way to hint arguments
    def __call__(self, f):
        if 'self' in inspect.getargspec(f).args:
            def new_f(s, *args, **kwargs):
                _args = []
                for service in self.get_services():
                    _args.append(service)
                for arg in args:
                    _args.append(arg)
                return f(s, *_args, **kwargs)
            return new_f
        else:
            def new_f(*args, **kwargs):
                _args = []
                for service in self.get_services():
                    _args.append(service)
                for arg in args:
                    _args.append(arg)
                return f(*_args, **kwargs)
            return new_f

    def get_services(self):
        """helper function to get injected services"""
        for id in self.ids:
            yield self.store.get_service(id, from_namespace=self.namespace)

service = Servicer
inject = Injector


"""
helpers.py

Simple helpers for adding and getting a service
"""

from di import ServiceStore

def get_service(id, ns='default'):
    """Gets a service based on the id."""
    return ServiceStore.get_service(id, from_namespace=ns)

def add_service(id, service, ns='default'):
    """Adds a service with an id.  Same as di.decorator.service('id')(service)."""
    ServiceStore.add_service(id, service, to_namespace=ns)



"""
The dependency injection module contains everything you need
to store and lookup services, and to inject them into functions,
including class functions.
"""
from collections import namedtuple
from base64 import b16encode
from exceptions import ServiceNotFound

Service = namedtuple('Service', 'ns, id, service')

class ServiceStore(object):
    """Storage for all your services."""
    services = {}

    @classmethod
    def add_service(cls, id, service, to_namespace='default'):
        """Add a service with an id to ServiceStore"""
        service_hash = b16encode(to_namespace + id)
        cls.services[service_hash] = Service(to_namespace, id, service)

    @classmethod
    def get_service(cls, id, from_namespace='default'):
        """Get a service by id, raises a ServiceNotFound exception if the service id is not found."""
        service_hash = b16encode(from_namespace + id)
        if service_hash in cls.services.keys():
            return cls.services[service_hash].service
        else:
            raise ServiceNotFound("Service " + id + " not found.")



from unittest import TestCase
from di.helpers import get_service, add_service
from di.exceptions import ServiceNotFound

class TestNamespacedServices(TestCase):
    def setUp(self):
        def nsfunc(arg): pass
        add_service('nsfunc', nsfunc, ns='ns')

    def test_get_service_without_namespace_does_not_get_nsfunc(self):
        self.assertRaises(ServiceNotFound, get_service, 'nsfunc')

    def test_get_service_with_namespace_returns_nsfunc(self):
        self.assertEqual('nsfunc', get_service('nsfunc', ns='ns').__name__)


"""
exceptions.py

Contains exception classes for CNSdi

ServiceNotFound is used by di.ServiceStore

All CNSdi exceptions require a message passed into the constructor.
"""

class BaseDIException(Exception):
    """Base exception class for CNSdi"""
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class ServiceNotFound(BaseDIException):
    """Used when a service cannot be found"""
    pass


from unittest import TestCase
from di.exceptions import ServiceNotFound
from di import ServiceStore

class TestExceptions(TestCase):

    def test_service_not_found_exception(self):
        self.assertRaises(ServiceNotFound, ServiceStore.get_service, 'does.not.exist')


from unittest import TestCase
from di.helpers import get_service, add_service
from di import ServiceStore

class O(object): pass

class TestHelpers(TestCase):
    def setUp(self):
        ServiceStore.add_service('object', O)

    def test_get_service_returns_class(self):
        service = get_service('object')
        self.assertEqual(O, service, "Incorrect service found.")

    def test_add_service(self):
        add_service('O', O)
        self.assertEqual(O, ServiceStore.get_service('O'), "Incorrect service found.")


from unittest import TestCase
from di.decorator import service, inject

@service('a')
class A(object): pass

@inject(['a'])
def partial(a, b):
    return b

class TestPartialInjection(TestCase):
    def test_injected_function_takes_argument_properly(self):
        self.assertEqual('b', partial('b'))